@@ -17,7 +17,11 @@ from os.path import join as joinpath
 from os import sep as slash
 from os import listdir, utime, stat
 
-from stat import S_ISREG, S_ISDIR
+from stat import S_ISREG, S_ISDIR, S_IMODE, S_ISBLK, S_ISSOCK
+from stat import S_IRUSR, S_IWUSR, S_IXUSR
+from stat import S_IRGRP, S_IWGRP, S_IXGRP
+from stat import S_IROTH, S_IWOTH, S_IXOTH
+
 
 # Please keep this as light as possible on other Twisted imports; many, many
 # things import this module, and it would be good if it could easily be
@@ -30,6 +34,8 @@ from twisted.python.win32 import ERROR_FILE_NOT_FOUND, ERROR_PATH_NOT_FOUND
 from twisted.python.win32 import ERROR_INVALID_NAME, ERROR_DIRECTORY, O_BINARY
 from twisted.python.win32 import WindowsError
 
+from twisted.python.util import FancyEqMixin
+
 _CREATE_FLAGS = (os.O_EXCL |
                  os.O_CREAT |
                  os.O_RDWR |
@@ -246,6 +252,15 @@ class _PathHelper:
 
 
     def sibling(self, path):
+        """
+        Return a L{FilePath} with the same directory as this instance but with a
+        basename of C{path}.
+
+        @param path: The basename of the L{FilePath} to return.
+        @type path: C{str}
+
+        @rtype: L{FilePath}
+        """
         return self.parent().child(path)
 
 
@@ -329,6 +344,95 @@ class _PathHelper:
 
 
 
+class RWX(FancyEqMixin, object):
+    """
+    A class representing read/write/execute permissions for a single user
+    category (i.e. user/owner, group, or other/world).  Instantiate with
+    three boolean values: readable? writable? executable?.
+
+    @type read: C{bool}
+    @ivar read: Whether permission to read is given
+
+    @type write: C{bool}
+    @ivar write: Whether permission to write is given
+
+    @type execute: C{bool}
+    @ivar execute: Whether permission to execute is given
+
+    @since: 11.1
+    """
+    compareAttributes = ('read', 'write', 'execute')
+    def __init__(self, readable, writable, executable):
+        self.read = readable
+        self.write = writable
+        self.execute = executable
+
+
+    def __repr__(self):
+        return "RWX(read=%s, write=%s, execute=%s)" % (
+            self.read, self.write, self.execute)
+
+
+    def shorthand(self):
+        """
+        Returns a short string representing the permission bits.  Looks like
+        part of what is printed by command line utilities such as 'ls -l'
+        (e.g. 'rwx')
+        """
+        returnval = ['r', 'w', 'x']
+        i = 0
+        for val in (self.read, self.write, self.execute):
+            if not val:
+                returnval[i] = '-'
+            i += 1
+        return ''.join(returnval)
+
+
+
+class Permissions(FancyEqMixin, object):
+    """
+    A class representing read/write/execute permissions.  Instantiate with any
+    portion of the file's mode that includes the permission bits.
+
+    @type user: L{RWX}
+    @ivar user: User/Owner permissions
+
+    @type group: L{RWX}
+    @ivar group: Group permissions
+
+    @type other: L{RWX}
+    @ivar other: Other/World permissions
+
+    @since: 11.1
+    """
+
+    compareAttributes = ('user', 'group', 'other')
+
+    def __init__(self, statModeInt):
+        self.user, self.group, self.other = (
+            [RWX(*[statModeInt & bit > 0 for bit in bitGroup]) for bitGroup in
+             [[S_IRUSR, S_IWUSR, S_IXUSR],
+              [S_IRGRP, S_IWGRP, S_IXGRP],
+              [S_IROTH, S_IWOTH, S_IXOTH]]]
+        )
+
+
+    def __repr__(self):
+        return "[%s | %s | %s]" % (
+            str(self.user), str(self.group), str(self.other))
+
+
+    def shorthand(self):
+        """
+        Returns a short string representing the permission bits.  Looks like
+        what is printed by command line utilities such as 'ls -l'
+        (e.g. 'rwx-wx--x')
+        """
+        return "".join(
+            [x.shorthand() for x in (self.user, self.group, self.other)])
+
+
+
 class FilePath(_PathHelper):
     """
     I am a path on the filesystem that only permits 'downwards' access.
@@ -379,16 +483,36 @@ class FilePath(_PathHelper):
     path = None
 
     def __init__(self, path, alwaysCreate=False):
+        """
+        Convert a path string to an absolute path if necessary and initialize
+        the L{FilePath} with the result.
+        """
         self.path = abspath(path)
         self.alwaysCreate = alwaysCreate
 
     def __getstate__(self):
+        """
+        Support serialization by discarding cached L{os.stat} results and
+        returning everything else.
+        """
         d = self.__dict__.copy()
         if d.has_key('statinfo'):
             del d['statinfo']
         return d
 
+
     def child(self, path):
+        """
+        Create and return a new L{FilePath} representing a path contained by
+        C{self}.
+
+        @param path: The base name of the new L{FilePath}.  If this contains
+            directory separators or parent references it will be rejected.
+        @type path: C{str}
+
+        @raise InsecurePath: If the result of combining this path with C{path}
+            would result in a path which is not a direct child of this path.
+        """
         if platform.isWindows() and path.count(":"):
             # Catch paths like C:blah that don't have a slash
             raise InsecurePath("%r contains a colon." % (path,))
@@ -717,9 +841,25 @@ class FilePath(_PathHelper):
         return int(st.st_gid)
 
 
+    def getPermissions(self):
+        """
+        Returns the permissions of the file.  Should also work on Windows,
+        however, those permissions may not what is expected in Windows.
+
+        @return: the permissions for the file
+        @rtype: L{Permissions}
+        @since: 11.1
+        """
+        st = self.statinfo
+        if not st:
+            self.restat()
+            st = self.statinfo
+        return Permissions(S_IMODE(st.st_mode))
+
+
     def exists(self):
         """
-        Check if the C{path} exists.
+        Check if this L{FilePath} exists.
 
         @return: C{True} if the stats of C{path} can be retrieved successfully,
             C{False} in the other cases.
@@ -736,6 +876,10 @@ class FilePath(_PathHelper):
 
 
     def isdir(self):
+        """
+        @return: C{True} if this L{FilePath} refers to a directory, C{False}
+            otherwise.
+        """
         st = self.statinfo
         if not st:
             self.restat(False)
@@ -744,7 +888,12 @@ class FilePath(_PathHelper):
                 return False
         return S_ISDIR(st.st_mode)
 
+
     def isfile(self):
+        """
+        @return: C{True} if this L{FilePath} points to a regular file (not a
+            directory, socket, named pipe, etc), C{False} otherwise.
+        """
         st = self.statinfo
         if not st:
             self.restat(False)
@@ -753,32 +902,96 @@ class FilePath(_PathHelper):
                 return False
         return S_ISREG(st.st_mode)
 
+
+    def isBlockDevice(self):
+        """
+        Returns whether the underlying path is a block device.
+
+        @return: C{True} if it is a block device, C{False} otherwise 
+        @rtype: C{bool}
+        @since: 11.1
+        """
+        st = self.statinfo
+        if not st:
+            self.restat(False)
+            st = self.statinfo
+            if not st:
+                return False
+        return S_ISBLK(st.st_mode)
+
+
+    def isSocket(self):
+        """
+        Returns whether the underlying path is a socket.
+
+        @return: C{True} if it is a socket, C{False} otherwise 
+        @rtype: C{bool}
+        @since: 11.1
+        """
+        st = self.statinfo
+        if not st:
+            self.restat(False)
+            st = self.statinfo
+            if not st:
+                return False
+        return S_ISSOCK(st.st_mode)
+
+
     def islink(self):
+        """
+        @return: C{True} if this L{FilePath} points to a symbolic link.
+        """
         # We can't use cached stat results here, because that is the stat of
         # the destination - (see #1773) which in *every case* but this one is
         # the right thing to use.  We could call lstat here and use that, but
         # it seems unlikely we'd actually save any work that way.  -glyph
         return islink(self.path)
 
+
     def isabs(self):
+        """
+        @return: C{True}, always.
+        """
         return isabs(self.path)
 
+
     def listdir(self):
+        """
+        List the base names of the direct children of this L{FilePath}.
+
+        @return: a C{list} of C{str} giving the names of the contents of the
+            directory this L{FilePath} refers to.  These names are relative to
+            this L{FilePath}.
+
+        @raise: Anything the platform C{os.listdir} implementation might raise
+            (typically OSError).
+        """
         return listdir(self.path)
 
+
     def splitext(self):
+        """
+        @return: tuple where the first item is the filename and second item is
+            the file extension. See Python docs for C{os.path.splitext}
+        """
         return splitext(self.path)
 
+
     def __repr__(self):
         return 'FilePath(%r)' % (self.path,)
 
+
     def touch(self):
+        """
+        Update the access and modified times of the L{FilePath}'s file.
+        """
         try:
             self.open('a').close()
         except IOError:
             pass
         utime(self.path, None)
 
+
     def remove(self):
         """
         Removes the file or directory that is represented by self.  If
@@ -813,13 +1026,30 @@ class FilePath(_PathHelper):
         path = self.path[-1] == '/' and self.path + pattern or slash.join([self.path, pattern])
         return map(self.clonePath, glob.glob(path))
 
+
     def basename(self):
+        """
+        @return: The final component of the L{FilePath}'s path (Everything after
+            the final path separator).
+        @rtype: C{str}
+        """
         return basename(self.path)
 
+
     def dirname(self):
+        """
+        @return: All of the components of the L{FilePath}'s path except the last
+            one (everything up to the final path separator).
+        @rtype: C{str}
+        """
         return dirname(self.path)
 
+
     def parent(self):
+        """
+        @return: A L{FilePath} representing the path which directly contains
+            this L{FilePath}.
+        """
         return self.clonePath(self.dirname())
 
 
@@ -888,14 +1118,25 @@ class FilePath(_PathHelper):
             return NotImplemented
         return cmp(self.path, other.path)
 
+
     def createDirectory(self):
+        """
+        Create the directory the L{FilePath} refers to.
+
+        @see: L{makedirs}
+
+        @raise OSError: If the directory cannot be created.
+        """
         os.mkdir(self.path)
 
+
     def requireCreate(self, val=1):
         self.alwaysCreate = val
 
+
     def create(self):
-        """Exclusively create a file, only if this file previously did not exist.
+        """
+        Exclusively create a file, only if this file previously did not exist.
         """
         fdint = os.open(self.path, _CREATE_FLAGS)
 
@@ -937,6 +1178,8 @@ class FilePath(_PathHelper):
         """
         Copies self to destination.
 
+        If self doesn't exist, an OSError is raised.
+
         If self is a directory, this method copies its children (but not
         itself) recursively to destination - if destination does not exist as a
         directory, this method creates it.  If destination is a file, an
@@ -1000,6 +1243,8 @@ class FilePath(_PathHelper):
                     readfile.close()
             finally:
                 writefile.close()
+        elif not self.exists():
+            raise OSError(errno.ENOENT, "No such file or directory")
         else:
             # If you see the following message because you want to copy
             # symlinks, fifos, block devices, character devices, or unix
